// printmatrix.cpp
// by Glenn G. Chappell
// October 2003
//
// For CS 381
// Getting and printing an OpenGL transformation matrix

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <iomanip>
using std::fixed;
using std::setprecision;
using std::setw;
#include <string>
using std::string;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff - includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels
const double neardist = 1.;   // Near & far clipping distances
const double fardist = 10.;

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// Scene
double angle = 0.;            // Rotation angle for object (degrees)
const double rotspeed = 1.;   // Rotation speed (degrees/frame)
bool rotating = false;        // True if currently rotating

// Matrix
GLdouble mvmatrixd[16];       // Holds the current model/view matrix
                              // Set in function display


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// drawsquare
// Draws a square
void drawsquare()
{
   glBegin(GL_QUADS);
      glVertex3d(-1., -1., 0.);
      glVertex3d( 1., -1., 0.);
      glVertex3d( 1.,  1., 0.);
      glVertex3d(-1.,  1., 0.);
   glEnd();
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   
   glPushMatrix();
      glTranslated(0., 0., -3.);  // Viewing transformation
      glRotated(angle, 0,1,0);

      glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrixd);
      // The above puts the current m/v matrix into mvmatrixd.

      // To do the reverse, and set the current m/v matrix equal
      //  to the contents of mvmatrixd, do "glLoadMatrixd(mvmatrixd);".
      //  More usefully, to multiply the contents of mvmatrixd by the
      //  current m/v matrix, do "glMultMatrixd(mvmatrixd);". Note that,
      //  for these last two to work, the matrix mode must have already
      //  been set to GL_MODELVIEW.

      glPushMatrix();
         glTranslated(0., 0., -1.);
         glColor3d(0.9, 0.1, 0.1);
         drawsquare();
      glPopMatrix();
      glPushMatrix();
         glTranslated(0., 0., 1.);
         glColor3d(0.1, 0.1, 0.9);
         drawsquare();
      glPopMatrix();

   glPopMatrix();

   // Text documentation
   glDisable(GL_DEPTH_TEST);  // No depth test for text
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();  // Save old projection matrix
      glLoadIdentity();
      gluOrtho2D(-1., 1., -1., 1.);  // old faithful
      glColor3d(0., 0., 0.);
      printbitmap("Printing a transformation matrix", -0.9, 0.9);
      printbitmap("Esc  Quit", -0.9, 0.7);
      printbitmap("R    Toggle rotation", -0.9, 0.6);
      printbitmap("Z    Reset view", -0.9, 0.5);
      printbitmap("P    Print the model/view matrix", -0.9, 0.4);
      printbitmap("     (in the text window)", -0.9, 0.3);
   glPopMatrix();  // Restore old projection matrix
   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
   glEnable(GL_DEPTH_TEST);  // Back to "normal" for 3-D drawing

   glutSwapBuffers();
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
   glViewport(0, 0, w, h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   // We set up coordinate system so that aspect ratios are always correct,
   //  and the region from -1..1 in x & y always just fits in the viewport.
   if (w > h)
   {
      glFrustum(-double(w)/h, double(w)/h, -1., 1., neardist, fardist);
      // You may want to use gluPerspective here
   }
   else
   {
      glFrustum(-1., 1., -double(h)/w, double(h)/w, neardist, fardist);
   }

   glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Update the scene
   if (rotating)
   {
      angle += rotspeed;
      glutPostRedisplay();
   }
}


// printmatrix
// Prints the given matrix
// Assumes column-major format, as in OpenGL
void printmatrix(GLdouble thematrix[16])
{
   int i, j;  // MS-Vis C++ 6.0 sometimes chokes on var's
              //  declared in for's.

   for (i=0; i<4; ++i)     // i = row
   {
      for (j=0; j<4; ++j)  // j = column
      {
         cout << fixed << setprecision(4) << setw(7)
               << thematrix[j*4+i] << " ";
      }
      cout << endl;
   }
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 'r':     // R: toggle rotation
   case 'R':
      rotating = !rotating;
      glutPostRedisplay();
      break;
   case 'p':     // P: print model/view matrix
   case 'P':
      // Matrix was already placed in mvmatrixd by the display function.
      cout << endl;
      cout << "Model/view matrix:" << endl;
      printmatrix(mvmatrixd);
      break;
   case 'z':     // Z: reset the view
   case 'Z':
      angle = 0.;
      rotating = false;
      glutPostRedisplay();
      break;
   }
}


// init
// Initialization
// Called by main after window creation
void init()
{
   // General
   glClearColor(0.7, 0.8, 0.8, 0.0);  // Light blue-gray background
   glEnable(GL_DEPTH_TEST);           // 3-D so we need HSR
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Printing a Transformation Matrix");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutIdleFunc(idle);
   glutKeyboardFunc(keyboard);

   // Do something
   glutMainLoop();

   return 0;
}
